# encoding: utf-8

module Verifier

  # Verification runner
  #
  # @api private
  class Verification

    # Initializes a verification object
    #
    # @param [Object] object
    #   the object whose method should be verified
    # @param [Symbol] variable
    #   the name of the variable to be verified
    # @param [Symbol] name
    #   the name of verification
    # @param [Proc] block
    #
    # @return [undefined]
    def initialize(object, variable, name, &block)
      fail SyntaxError.new "No block given" unless block_given?
      @object, @variable, @name, @block = object, variable, name, block
    end

    # Runs a verification
    #
    # @yield block in the scope of a value returned by the {#variable}
    #
    # @raise  [Verifier::MethodNotDefined]
    #   if the {#variable} not defined by {#object}
    # @raise  [Verifier::MethodFails]
    #   it calling the {#variable} raises a +StandardError+
    # @raise  [Verifier::Invalid]
    #   if the block fails or returns a value whose negation is truthy
    #
    # @return [undefined]
    def run
      check_method_existence
      check_condition
    end

    # Constructs and runs a verification object
    #
    # @param  (see #initialize)
    #
    # @raise  (see #run)
    #
    # @return (see #run)
    def self.run(object, variable, name, &block)
      send(:new, object, variable, name, &block).run
    end

    # @note
    #   The class has a private constructor. Use {.run} method instead.
    private_class_method :new

    private

    attr_reader :object, :variable, :name, :block

    def check_method_existence
      fail(undefined) unless object.respond_to?(variable)
    end

    def value
      @value ||= object.send(variable)
    rescue
      raise method_fails
    end

    def check_condition
      fail unless value.instance_eval(&block)
    rescue
      raise invalid
    end

    # Exceptions to be raised

    def undefined
      MethodNotDefined.new translate(
        :undefined,
        scope:   scope.short,
        value:   variable,
        default: "Method not defined: #{ method_name }"
      )
    end

    def method_fails
      MethodFails.new translate(
        :fails,
        scope:   scope.short,
        value:   variable,
        default: "Method fails: #{ method_name }"
      )
    end

    def invalid
      Invalid.new translate(
        name,
        scope:   scope.full,
        value:   value,
        default: default_message
      )
    end

    def default_message
      "#{ method_name } verification failed: #{ value.inspect } not #{ name }"
    end

    # Translation helpers

    def method_name
      @method_name ||= "#{ object.class.name }#{ scope.tag }#{ variable }"
    end

    def scope
      @scope ||= Scope.new(object, variable)
    end

    def translate(name, options = {})
      I18n.t(name, options)
    end
  end
end
